namespace X3DTK {
namespace MESH {

template<class MData, class VData, class EData, class FData, bool readOnly>
EdgesView<MData, VData, EData, FData, readOnly>::EdgesView() 
: View("edges"), _modified(true), _modifiedSelection(true), _displayList(0), _selectionDisplayList(0), _selectedMesh(0), _selectedEdge(0), _selectedTransform(SFMatrix34f::identity)
{
  _edgesDrawer = X3DProcessor::getInstanceOf<EdgesDrawer<MData, VData, EData, FData, readOnly> >();
}

template<class MData, class VData, class EData, class FData, bool readOnly>
EdgesView<MData, VData, EData, FData, readOnly>::~EdgesView()
{
  glDeleteLists(_displayList, 1);
  glDeleteLists(_selectionDisplayList, 1);
}

template<class MData, class VData, class EData, class FData, bool readOnly>
void EdgesView<MData, VData, EData, FData, readOnly>::receiveNotification()
{
  _modified = true;
  _modifiedSelection = true;
  glDeleteLists(_displayList, 1);
  glDeleteLists(_selectionDisplayList, 1);
  _displayList = glGenLists(1);
  _selectionDisplayList = glGenLists(1);
  _selectedMesh = 0;
  _selectedEdge = 0;
  _selectedTransform = SFMatrix34f::identity;
}

template<class MData, class VData, class EData, class FData, bool readOnly>
void EdgesView<MData, VData, EData, FData, readOnly>::display()
{
  if (_modified)
  {
    glNewList(_displayList, GL_COMPILE_AND_EXECUTE);
    _edgesDrawer->draw(getScene());
    glEndList();
    _modified = false;
  }  
  else
    glCallList(_displayList);
}

template<class MData, class VData, class EData, class FData, bool readOnly>
void EdgesView<MData, VData, EData, FData, readOnly>::initSelect()
{
  _selectedMesh = 0;
  _selectedEdge = 0;
  _selectedTransform = SFMatrix34f::identity;
}

template<class MData, class VData, class EData, class FData, bool readOnly>
void EdgesView<MData, VData, EData, FData, readOnly>::select()
{
  const int NB_HITS_MAX = 32768;
  // Prepare the selection mode
  static GLuint hits[NB_HITS_MAX];

  glSelectBuffer(NB_HITS_MAX, hits);
  glRenderMode(GL_SELECT);
  glInitNames();

  if (_modifiedSelection)
  {
    glNewList(_selectionDisplayList, GL_COMPILE_AND_EXECUTE);
    _edgesDrawer->draw(getScene(), true);
    glEndList();
    _modifiedSelection = false;
  }  
  else
    glCallList(_selectionDisplayList);

  glFlush();

  // Get the results
  GLint nb_hits = glRenderMode(GL_RENDER);

  // Too many elements in selection buffer
  if (nb_hits < 0)
    return;

  // No selection
  if (nb_hits == 0)
  {
    _selectedMesh = 0;
    _selectedEdge = 0;
    _selectedTransform = SFMatrix34f::identity;
    return;
  }
 
  //interpreting the results.
  _zMin = hits[1];
  int selectedMesh(hits[3]);
  int selectedId(hits[4]);

  for (int i = 1; i < nb_hits; ++i)
  {
    // Prefer vertices over edges and edges over faces
    if (hits[i*5 + 1] < _zMin)
    {
      _zMin = hits[i*5 + 1];
      selectedMesh = hits[i*5 + 3];
      selectedId = hits[i*5 + 4];
    }
  }  

  _selectedMesh = _edgesDrawer->getMesh(selectedMesh);
  _selectedEdge = _selectedMesh->getEdges()[selectedId];
  _selectedTransform = _edgesDrawer->getMatrix(selectedMesh);
}

template<class MData, class VData, class EData, class FData, bool readOnly>
unsigned int EdgesView<MData, VData, EData, FData, readOnly>::getValue() const
{
  if (_selectedEdge != 0)
    return _zMin;
  return std::numeric_limits<unsigned int>::max();  
}

template<class MData, class VData, class EData, class FData, bool readOnly>
typename Mesh<MData, VData, EData, FData, readOnly>::SFEdge *EdgesView<MData, VData, EData, FData, readOnly>::getSelectedEdge() const 
{
  return _selectedEdge;
}

template<class MData, class VData, class EData, class FData, bool readOnly>
Mesh<MData, VData, EData, FData, readOnly> *EdgesView<MData, VData, EData, FData, readOnly>::getSelectedMesh() const 
{
  return _selectedMesh;
}

template<class MData, class VData, class EData, class FData, bool readOnly>
const SFMatrix34f &EdgesView<MData, VData, EData, FData, readOnly>::getSelectedTransform() const 
{
  return _selectedTransform;
}

}
}
